# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration for REANA client."""

from __future__ import absolute_import, print_function

import pytest


@pytest.fixture()
def create_yaml_workflow_schema():
    """Return dummy yaml workflow schema."""
    reana_yaml_schema = \
        '''
        version: 0.4.0
        inputs:
          files:
            - code/helloworld.py
            - inputs/names.txt
          parameters:
            sleeptime: 2
            inputfile: inputs/names.txt
            helloworld: code/helloworld.py
            outputfile: outputs/greetings.txt
        outputs:
          files:
           - outputs/greetings.txt
        workflow:
          type: serial
          specification:
            steps:
              - environment: 'python:2.7'
                commands:
                  - python "${helloworld}" --sleeptime ${sleeptime} \
                  --inputfile "${inputfile}" --outputfile "${outputfile}"
        '''
    return reana_yaml_schema



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client files tests."""

import hashlib
import json
import os

from click.testing import CliRunner
from mock import Mock, patch
from pytest_reana.test_utils import make_mock_api_client

from reana_client.cli import cli


def test_list_files_server_not_reachable():
    """Test list workflow workspace files when not connected to any cluster."""
    reana_token = '000000'
    message = 'REANA client is not connected to any REANA cluster.'
    runner = CliRunner()
    result = runner.invoke(cli, ['ls', '-t', reana_token])
    assert result.exit_code == 1
    assert message in result.output


def test_list_files_server_no_token():
    """Test list workflow workspace files when access token is not set."""
    message = 'Please provide your access token'
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ['ls'])
    assert result.exit_code == 1
    assert message in result.output


def test_list_files_ok():
    """Test list workflow workspace files successfull."""
    status_code = 200
    response = [
        {
            "last-modified": "string",
            "name": "string",
            "size": 0
        }
    ]
    env = {'REANA_SERVER_URL': 'localhost'}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(
                cli, ['ls', '-t', reana_token, '--workflow', 'mytest.1',
                      '--json'])
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert json_response[0]['name'] in response[0]['name']


def test_list_files_url():
    """Test list workflow workspace files' urls."""
    status_code = 200
    response = [
        {
            "last-modified": "string",
            "name": "string",
            "size": 0
        }
    ]
    env = {'REANA_SERVER_URL': 'localhost'}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    workflow_name = 'mytest'
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(
                cli, ['ls', '-t', reana_token, '--workflow', workflow_name,
                      '--url'])
            assert result.exit_code == 0
            assert workflow_name in result.output
            assert response[0]['name'] in result.output


def test_download_file():
    """Test file downloading."""
    status_code = 200
    response = 'Content of file to download'
    env = {'REANA_SERVER_URL': 'localhost'}
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_http_response.content = str(response).encode()
    mock_requests = Mock()
    mock_requests.get = Mock(return_value=mock_http_response)

    reana_token = '000000'
    response_md5 = hashlib.md5(response.encode('utf-8')).hexdigest()
    file = 'dummy_file.txt'
    message = 'File {0} downloaded to'.format(file)
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.requests", mock_requests):
            result = runner.invoke(
                cli, ['download', '-t', reana_token, '--workflow', 'mytest.1',
                      file]
            )
            assert result.exit_code == 0
            assert os.path.isfile(file) is True
            file_md5 = hashlib.md5(open(file, 'rb').read()).hexdigest()
            assert file_md5 == response_md5
            assert message in result.output
            os.remove(file)


def test_upload_file(create_yaml_workflow_schema):
    """Test upload file."""
    reana_token = '000000'
    file = 'file.txt'
    env = {'REANA_SERVER_URL': 'http://localhost'}
    message = 'was successfully uploaded.'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.requests.post") as post_request:
            with runner.isolated_filesystem():
                with open(file, 'w') as f:
                    f.write('test')
                with open('reana.yaml', 'w') as reana_schema:
                    reana_schema.write(create_yaml_workflow_schema)
                result = runner.invoke(
                    cli, ['upload', '-t', reana_token, '--workflow',
                          'mytest.1', file]
                )
                post_request.assert_called_once()
                assert result.exit_code == 0
                assert message in result.output


def test_delete_file():
    """Test delete file."""
    status_code = 200
    reana_token = '000000'
    filename1 = 'file1'
    filename2 = 'problematic_file'
    filename2_error_message = '{} could not be deleted.'.format(filename2)
    response = {'deleted': {filename1: {'size': 19}},
                'failed': {filename2: {'error': filename2_error_message}}}
    message1 = 'file1 was successfully deleted'
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_http_response.raw_bytes = str(response).encode()
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            with runner.isolated_filesystem():
                result = runner.invoke(
                    cli,
                    ['rm', '-t', reana_token,
                     '--workflow', 'mytest.1', filename1]
                )
                assert result.exit_code == 0
                assert message1 in result.output
                assert filename2_error_message in result.output


def test_delete_non_existing_file():
    """Test delete non existing file."""
    status_code = 200
    reana_token = '000000'
    filename = 'file11'
    response = {'deleted': {}, 'failed': {}}
    message = '{} did not match any existing file.'.format(filename)
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_http_response.raw_bytes = str(response).encode()
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            with runner.isolated_filesystem():
                result = runner.invoke(
                    cli,
                    ['rm', '-t', reana_token, '--workflow', 'mytest.1',
                     filename]
                )
                assert result.exit_code == 0
                assert message in result.output


def test_move_file_running_workflow():
    """Test move files when workflow is running."""
    status_code = 200
    reana_token = '000000'
    src_file = 'file11'
    target = 'file2'
    response = {"status": "running",
                "logs": "",
                "name": "mytest.1"}
    message = 'File(s) could not be moved for running workflow'
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_http_response.raw_bytes = str(response).encode()
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(
                cli,
                ['mv', '-t', reana_token, '--workflow', 'mytest.1',
                 src_file, target]
            )
            assert result.exit_code == 1
            assert message in result.output



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client configuration."""

import os

import pkg_resources

reana_yaml_default_file_path = './reana.yaml'  # e.g. `./.reana.yaml`
"""REANA specification file default location."""

reana_yaml_schema_file_path = pkg_resources.resource_filename(
        __name__,
        'schemas/reana_analysis_schema.json')
"""REANA specification schema location."""

default_user = '00000000-0000-0000-0000-000000000000'
"""Default user to use when submitting workflows to REANA Server."""

ERROR_MESSAGES = {
        'missing_access_token':
        'Please provide your access token by using'
        ' the -t/--access-token flag, or by setting the'
        ' REANA_ACCESS_TOKEN environment variable.'
}

JSON = 'json'
"""Json output format."""

TIMECHECK = 5
"""Time between workflow status check."""

URL = 'url'
"""Url output format."""

WORKFLOW_ENGINES = ['serial', 'cwl', 'yadage']
"""Supported workflow engines."""



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2019 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client secrets tests."""

import pytest
from bravado.exception import HTTPError
from click.testing import CliRunner
from mock import Mock, patch
from pytest_reana.test_utils import make_mock_api_client

from reana_client.cli import cli


def test_secrets_list_server_not_reachable():
    """Test list secrets when not connected to any cluster."""
    message = 'REANA client is not connected to any REANA cluster.'
    reana_token = '000000'
    runner = CliRunner()
    result = runner.invoke(cli, ['secrets-list', '-t', reana_token])
    assert result.exit_code == 1
    assert message in result.output


def test_secrets_list_server_no_token():
    """Test list secrets when access token is not set."""
    message = 'Please provide your access token'
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ['secrets-list'])
    assert result.exit_code == 1
    assert message in result.output


def test_secrets_list_ok():
    """Test list secrets successfull."""
    status_code = 200
    response = [
        {
            "name": "password",
            "type": "env"
        }
    ]
    env = {'REANA_SERVER_URL': 'localhost'}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(
                cli, ['secrets-list', '-t', reana_token])
            assert result.exit_code == 0
            assert "password" in result.output
            assert "env" in result.output


@pytest.mark.parametrize(
    'secret',
    ['USER=reanauser', 'USER=reana=user'])
def test_secrets_add(secret):
    """Test secrets add."""
    status_code = 201
    reana_token = '000000'
    secret_file = 'file.txt'
    response = [secret_file]
    env = {'REANA_SERVER_URL': 'localhost'}
    message = 'were successfully uploaded.'
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_http_response.raw_bytes = str(response).encode()
    mock_response = response
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            with runner.isolated_filesystem():
                with open(secret_file, 'w') as f:
                    f.write('test')
                result = runner.invoke(
                    cli, ['secrets-add',
                          '-t', reana_token,
                          '--file', secret_file,
                          '--env', secret]
                )
                assert result.exit_code == 0
                assert message in result.output


@pytest.mark.parametrize(
    'secret',
    ['wrongformat', 'PASS:123'])
def test_secrets_add_wrong_format(secret):
    """Test adding secrets with wrong format."""
    reana_token = '000000'
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    message = 'For literal strings use "SECRET_NAME=VALUE" format'

    result = runner.invoke(
        cli, ['secrets-add', '-t', reana_token,
              '--env', secret]
    )
    assert result.exit_code == 1
    assert message in result.output


def test_secrets_add_already_exist():
    """Test adding secrets when they already exist."""
    status_code = 409
    reana_token = '000000'
    env = {'REANA_SERVER_URL': 'localhost'}
    message = 'One of the secrets already exists. No secrets were added.'
    mock_http_response = Mock(
        status_code=status_code,
        reason='Conflict',
        json=Mock(return_value={'message': 'Conflict'}))
    rs_api_client_mock = Mock()
    rs_api_client_mock.api.add_secrets = Mock(
        side_effect=HTTPError(mock_http_response))
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                rs_api_client_mock):
                result = runner.invoke(
                    cli, ['secrets-add',
                          '-t', reana_token,
                          '--env', 'USER=reanauser']
                )
                assert message in result.output
                assert result.exit_code == 1


def test_secrets_delete():
    """Test secrets delete."""
    status_code = 200
    reana_token = '000000'
    secret_file = 'file.txt'
    response = ['file.txt']
    message = 'were successfully deleted.'
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
                result = runner.invoke(
                    cli, ['secrets-delete', '-t', reana_token, secret_file]
                )
                assert result.exit_code == 0
                assert message in result.output



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client workflow tests."""

import json

import pytest
import yaml
from click.testing import CliRunner
from mock import Mock, patch
from pytest_reana.test_utils import make_mock_api_client
from reana_commons.config import INTERACTIVE_SESSION_TYPES

from reana_client.api.client import create_workflow_from_json
from reana_client.cli import cli
from reana_client.utils import get_workflow_status_change_msg


def test_workflows_server_not_connected():
    """Test workflows command when server is not connected."""
    runner = CliRunner()
    reana_token = '000000'
    result = runner.invoke(cli, ['list', '-t', reana_token])
    message = 'REANA client is not connected to any REANA cluster.'
    assert message in result.output
    assert result.exit_code == 1


def test_workflows_no_token():
    """Test workflows command when token is not set."""
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ['list'])
    message = 'Please provide your access token by using the -t'
    assert result.exit_code == 1
    assert message in result.output


def test_workflows_server_ok():
    """Test workflows command when server is reachable."""
    response = [
        {
            "status": "running",
            "created": "2018-06-13T09:47:35.66097",
            "user": "00000000-0000-0000-0000-000000000000",
            "name": "mytest.1",
            "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        }
    ]
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost', 'REANA_WORKON': 'mytest.1'}
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(cli, ['list', '-t', reana_token])
            message = 'RUN_NUMBER'
            assert result.exit_code == 0
            assert message in result.output


def test_workflows_sorting():
    """Test workflows sorting."""
    response = [
        {
            "status": "running",
            "created": "2018-06-13T09:47:35.66097",
            "user": "00000000-0000-0000-0000-000000000000",
            "name": "mytest.1",
            "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
            "progress": {
                "run_started_at": "2018-06-13T09:47:40.28223",
                "run_finished_at": "2018-06-13T10:30:03.70303"
            }
        },
        {
            "status": "running",
            "created": "2018-06-13T09:55:35.66097",
            "user": "00000000-0000-0000-0000-000000000000",
            "name": "mytest.2",
            "id": "256b25f4-4cfb-4684-b7a8-73872ef455a2",
        }
    ]
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost', 'REANA_WORKON': 'mytest.1'}
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(
                cli,
                ['list', '-t', reana_token, '--sort', 'run_number'])
            message = (
                'mytest   2            2018-06-13T09:55:35.66097   -'
                '                           -'
                '                           running\n'
                'mytest   1            2018-06-13T09:47:35.66097   '
                '2018-06-13T09:47:40.28223   2018-06-13T10:30:03.70303'
                '   running')
            assert result.exit_code == 0
            assert message in result.output


def test_workflows_sessions():
    """Test list command for getting interactive sessions."""
    response = [
        {
            'created': '2019-03-19T14:37:58',
            'id': '29136cd0-b259-4d48-8c1e-afe3572df408',
            'name': 'workflow.1',
            'session_type': 'jupyter',
            'session_uri': '/29136cd0-b259-4d48-8c1e-afe3572df408',
            'size': '0',
            'status': 'created',
            'user': '00000000-0000-0000-0000-000000000000'
        }
    ]
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost', 'REANA_WORKON': 'mytest.1'}
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(cli, ['list', '-t', reana_token,
                                         '--sessions'])
            message = 'RUN_NUMBER'
            assert result.exit_code == 0
            assert message in result.output


def test_workflows_valid_json():
    """Test workflows command with --json and -v flags."""
    response = [
        {
            "status": "running",
            "created": "2018-06-13T09:47:35.66097",
            "user": "00000000-0000-0000-0000-000000000000",
            "name": "mytest.1",
            "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
            "size": "0K"
        }
    ]
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost'}
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(cli,
                                   ['list', '-v', '-t',
                                    reana_token, '--json'])
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert 'name' in json_response[0]
            assert 'run_number' in json_response[0]
            assert 'created' in json_response[0]
            assert 'status' in json_response[0]
            assert 'id' in json_response[0]
            assert 'user' in json_response[0]


def test_workflows_filter():
    """Test workflows command with --filter."""
    response = [
        {
            "status": "running",
            "created": "2018-06-13T09:47:35.66097",
            "user": "00000000-0000-0000-0000-000000000000",
            "name": "mytest.1",
            "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        },
        {
            "status": "failed",
            "created": "2018-06-14T09:47:35.66097",
            "user": "00000000-0000-0000-0000-000000000000",
            "name": "mytest.2",
            "id": "256b25f4-4cfb-4684-b7a8-73872ef455a2",
        }
    ]
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost'}
    reana_token = '000000'
    runner = CliRunner(env=env)
    filter_ = 'status=failed'
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(cli,
                                   ['list', '-t', reana_token, '--json',
                                    '--format="{}"'.format(filter_)])
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert 'status' in json_response[0]
            assert json_response[0]['status'] == 'failed'


def test_workflow_create_failed():
    """Test workflow create when creation fails."""
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ['create'])
    message = 'Error: Invalid value for "-f"'
    assert message in result.output
    assert result.exit_code == 2


def test_workflow_create_successful(create_yaml_workflow_schema):
    """Test workflow create when creation is successfull."""
    status_code = 201
    response = {
        "message": "The workflow has been successfully created.",
        "workflow_id": "cdcf48b1-c2f3-4693-8230-b066e088c6ac",
        "workflow_name": "mytest.1"
    }
    env = {'REANA_SERVER_URL': 'localhost'}
    reana_token = '000000'
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            with runner.isolated_filesystem():
                with open('reana.yaml', 'w') as f:
                    f.write(create_yaml_workflow_schema)
                result = runner.invoke(
                    cli,
                    ['create', '-t', reana_token, '--skip-validation']
                )
                assert result.exit_code == 0
                assert response["workflow_name"] in result.output


def test_workflow_create_not_valid_name(create_yaml_workflow_schema):
    """Test workflow create when creation is successfull."""
    env = {'REANA_SERVER_URL': 'localhost'}
    illegal_workflow_name = 'workflow.name'
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ['create', '-n', illegal_workflow_name])
    message = 'Workflow name {} contains illegal character "{}"'.format(
        illegal_workflow_name,
        '.'
    )
    assert message in result.output
    assert result.exit_code == 1


def test_create_workflow_from_json(create_yaml_workflow_schema):
    """Test create workflow from json specification."""
    status_code = 201
    response = {
        "message": "The workflow has been successfully created.",
        "workflow_id": "cdcf48b1-c2f3-4693-8230-b066e088c6ac",
        "workflow_name": "mytest.1"
    }
    env = {'REANA_SERVER_URL': 'localhost'}
    reana_token = '000000'
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    workflow_json = yaml.load(create_yaml_workflow_schema,
                              Loader=yaml.FullLoader)
    with patch.dict('os.environ', env):
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = create_workflow_from_json(
                workflow_json=workflow_json['workflow'],
                name=response['workflow_name'],
                access_token=reana_token,
                parameters=workflow_json['inputs'],
                workflow_engine='serial')
            assert response['workflow_name'] == result['workflow_name']
            assert response['message'] == result['message']


@pytest.mark.parametrize("status", [
    'created', 'running', 'finished', 'failed', 'deleted', 'stopped', 'queued'
])
def test_workflow_start_successful(status):
    """Test workflow start when creation is successfull."""
    workflow_name = "mytest.1"
    response = {
        "status": status,
        "message": "Server message",
        "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        "workflow_name": workflow_name,
        "user": "00000000-0000-0000-0000-000000000000"
    }
    status_code = 200
    reana_token = '000000'
    expected_message = get_workflow_status_change_msg(workflow_name, status)
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(
                cli,
                ['start', '-t', reana_token, '-w', response["workflow_name"]]
            )
            assert result.exit_code == 0
            assert expected_message in result.output


@pytest.mark.parametrize(
    "initial_status, final_status, exit_code",
    [('running', 'finished', 0), ('running', 'failed', 1),
        ('running', 'stopped', 1)])
def test_workflow_start_follow(initial_status, final_status, exit_code):
    """Test start workflow with follow flag."""
    workflow_name = "mytest.1"
    initial_reponse = {
        "status": initial_status,
        "message": "Server message",
        "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        "workflow_name": workflow_name,
        "user": "00000000-0000-0000-0000-000000000000"}
    final_reponse = {
        "status": final_status,
        "message": "Server message",
        "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        "workflow_name": workflow_name,
        "user": "00000000-0000-0000-0000-000000000000"}
    initial_expected_message = get_workflow_status_change_msg(
        workflow_name, initial_status)
    final_Expected_message = get_workflow_status_change_msg(
        workflow_name, final_status)
    status_code = 200
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_api_client = Mock()
    mock_start_workflow_result = Mock(
        return_value=(mock_http_response, mock_http_response))
    mock_api_client.api.start_workflow.return_value = \
        Mock(result=mock_start_workflow_result)
    mock_get_workflow_status_result = Mock(
        side_effect=[(initial_reponse, mock_http_response),
                     (final_reponse, mock_http_response)])
    mock_api_client.api.get_workflow_status.return_value = \
        Mock(result=mock_get_workflow_status_result)
    reana_token = '000000'
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                mock_api_client):
            result = runner.invoke(
                cli,
                ['start', '-t', reana_token, '-w', workflow_name, '--follow']
            )
            assert result.exit_code == exit_code
            assert initial_expected_message in result.output
            assert final_Expected_message in result.output


def test_workflows_validate(create_yaml_workflow_schema):
    """Test validation of REANA specifications file."""
    message = "is a valid REANA specification file"
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    with runner.isolated_filesystem():
        with open('reana.yaml', 'w') as f:
            f.write(create_yaml_workflow_schema)
        result = runner.invoke(
            cli,
            ['validate', '--file', 'reana.yaml'],
        )
        assert result.exit_code == 0
        assert message in result.output


def test_get_workflow_status_ok():
    """Test workflow status."""
    status_code = 200
    response = {
        'created': '2018-10-29T12:50:12',
        'id': '4e576cf9-a946-4346-9cde-7712f8dcbb3f',
        'logs': '',
        'name': 'workflow.5',
        'progress': {
            'current_command': None,
            'current_step_name': None,
            'failed': {'job_ids': [], 'total': 0},
            'finished': {'job_ids': [], 'total': 0},
            'run_started_at': '2018-10-29T12:51:04',
            'running': {'job_ids': [], 'total': 0},
            'total': {'job_ids': [], 'total': 1}
        },
        'status': 'running',
        'user': '00000000-0000-0000-0000-000000000000'
    }
    env = {'REANA_SERVER_URL': 'localhost'}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(
                cli,
                ['status', '-t', reana_token, '--json', '-v', '-w',
                 response['name']]
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert json_response[0]['name'] in response['name']


@patch('reana_client.cli.workflow.workflow_create')
@patch('reana_client.cli.workflow.upload_files')
@patch('reana_client.cli.workflow.workflow_start')
def test_run(workflow_start_mock,
             upload_file_mock,
             workflow_create_mock,
             create_yaml_workflow_schema):
    """Test run command, if wrapped commands are called."""
    reana_workflow_schema = "reana.yaml"
    env = {'REANA_SERVER_URL': 'localhost'}
    runner = CliRunner(env=env)
    reana_token = '000000'
    with runner.isolated_filesystem():
        with open(reana_workflow_schema, 'w') as f:
            f.write(create_yaml_workflow_schema)
        result = runner.invoke(
            cli,
            ['run', '-t', reana_token, '-f', reana_workflow_schema],
        )
    assert workflow_create_mock.called is True
    assert upload_file_mock.called is True
    assert workflow_start_mock.called is True


def test_workflow_input_parameters():
    """Test if not existing input parameters from CLI are applied."""
    status_code = 200
    response = {'id': 'd9304bdf-0d19-45d9-ae87-d5fd18059193',
                'name': 'workflow.19',
                'type': 'serial',
                'parameters': {'helloworld': 'code/helloworld.py',
                               'inputfile': 'data/names.txt',
                               'outputfile': 'results/greetings.txt',
                               'sleeptime': 2}}
    env = {'REANA_SERVER_URL': 'localhost'}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            parameter = "Debug"
            expected_message = '{0}, is not in reana.yaml'.format(parameter)
            result = runner.invoke(
                cli,
                ['start', '-t', reana_token, '-w workflow.19',
                 '-p {0}=True'.format(parameter)]
            )
            assert expected_message in result.output


@pytest.mark.parametrize(
    'interactive_session_type',
    INTERACTIVE_SESSION_TYPES +
    [pytest.param('wrong-interactive-type', marks=pytest.mark.xfail)])
def test_open_interactive_session(interactive_session_type):
    """Test opening an interactive session."""
    status_code = 200
    workflow_id = 'd9304bdf-0d19-45d9-ae87-d5fd18059193'
    response = {'path': '/{}'.format(workflow_id)}
    reana_server_url = 'http://localhost'
    env = {'REANA_SERVER_URL': reana_server_url}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            expected_message = '{reana_server_url}/{workflow_id}'.format(
                reana_server_url=reana_server_url, workflow_id=workflow_id)
            result = runner.invoke(cli, ['open', '-t', reana_token,
                                         '-w', workflow_id,
                                         interactive_session_type])
            assert expected_message in result.output


def test_close_interactive_session():
    """Test closing an interactive session."""
    status_code = 200
    workflow = "workflow.1"
    expected_message = "Interactive session for workflow {} " \
                       "was successfully closed\n".format(workflow)
    reana_server_url = 'http://localhost'
    env = {'REANA_SERVER_URL': reana_server_url}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = expected_message
    reana_token = '000000'
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):
            result = runner.invoke(cli, ['close', '-t', reana_token,
                                         '-w', workflow])
            assert expected_message == result.output



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client ping tests."""

from click.testing import CliRunner
from mock import Mock, patch
from pytest_reana.test_utils import make_mock_api_client

from reana_client.cli import cli


def test_ping_server_not_set():
    """Test ping when server is not set."""
    runner = CliRunner()
    result = runner.invoke(cli, ['ping'])
    message = 'REANA client is not connected to any REANA cluster.'
    assert message in result.output


def test_ping_server_not_reachable():
    """Test ping when server is set, but unreachable."""
    env = {'REANA_SERVER_URL': 'localhot'}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ['ping'])
    message = 'Could not connect to the selected'
    assert message in result.output


def test_ping_ok():
    """Test ping server is set and reachable."""
    env = {'REANA_SERVER_URL': 'localhost'}
    status_code = 200
    response = {"status": 200, "message": "OK"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
                "reana_client.api.client.current_rs_api_client",
                make_mock_api_client('reana-server')(mock_response,
                                                     mock_http_response)):

            result = runner.invoke(cli, ['ping'])
            message = 'Server is running'
            assert message in result.output



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Common click options."""

import functools
import json
import os
import shlex
import sys

import click

from reana_client.utils import workflow_uuid_or_name
from reana_client.config import ERROR_MESSAGES
from reana_commons.errors import MissingAPIClientConfiguration


def add_access_token_options(func):
    """Add access token related options to click commands."""
    @click.option('-t', '--access-token',
                  default=os.getenv('REANA_ACCESS_TOKEN', None),
                  callback=access_token_check,
                  help='Access token of the current user.')
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


def access_token_check(ctx, _, access_token):
    """Check if access token is present."""
    if not access_token:
        click.echo(
            click.style(ERROR_MESSAGES['missing_access_token'],
                        fg='red'), err=True)
        ctx.exit(1)
    else:
        return access_token


def check_connection(func):
    """Check if connected to any REANA cluster."""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from reana_client.utils import get_api_url
        api_url = get_api_url()
        if not api_url:
            click.secho(
                'REANA client is not connected to any REANA cluster.',
                fg='red', err=True)
            sys.exit(1)
        return func(*args, **kwargs)
    return wrapper


def add_workflow_option(func):
    """Add workflow related option to click commands."""
    @click.option('-w', '--workflow',
                  default=os.environ.get('REANA_WORKON', None),
                  callback=workflow_uuid_or_name,
                  help='Name or UUID of the workflow. Overrides value of '
                       'REANA_WORKON environment variable.')
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


def parse_parameters(_filter):
    """Return parsed filter parameters."""
    try:
        parsed_filters = []
        filters = ' '.join(_filter).replace(',', ' ')
        for item in shlex.split(filters):
            if '=' in item:
                filter_item = {
                    'column_name': item.split('=')[0].lower(),
                    'column_value': item.split('=')[1]}
            else:
                filter_item = {
                    'column_name': item.lower(),
                    'column_value': None}
            parsed_filters.append(filter_item)
        return parsed_filters
    except ValueError as e:
        click.echo(click.style('Wrong filter format \n{0}'
                               .format(e.message),
                               fg='red'), err=True)


def filter_data(parsed_filters, headers, tablib_data):
    """Return filtered data."""
    parsed_filters = \
        [i for i in parsed_filters if i['column_name'] in headers]
    column_headers = [i['column_name'] for i in parsed_filters] or None
    tablib_data = tablib_data.subset(rows=None,
                                     cols=column_headers)
    tablib_data = json.loads(tablib_data.export('json'))
    filtered_data = list(tablib_data)
    for item in filtered_data:
        for filter_ in parsed_filters:
            if (filter_['column_value'] is not None and
                    filter_['column_value'] != item[filter_['column_name']]):
                tablib_data.remove(item)
    return tablib_data, column_headers or []


def format_session_uri(reana_server_url, path, access_token):
    """Format interactive session URI."""
    return "{reana_server_url}{path}?token={access_token}".format(
        reana_server_url=reana_server_url,
        path=path, access_token=access_token)


def validate_workflow_name(ctx, _, workflow_name):
    """Validate workflow name."""
    not_allowed_characters = ['.']
    if workflow_name:
        for item in not_allowed_characters:
            if item in workflow_name:
                click.echo(
                    click.style(
                        'Workflow name {} contains illegal '
                        'character "{}""'.format(
                            workflow_name,
                            item
                        ),
                        fg='red'),
                    err=True)
                sys.exit(1)
    return workflow_name


class NotRequiredIf(click.Option):
    """Allow only one of two arguments to be missing."""

    def __init__(self, *args, **kwargs):
        """."""
        self.not_required_if = kwargs.pop('not_required_if')
        assert self.not_required_if, "'not_required_if' parameter required"
        super(NotRequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        """Overwritten click method."""
        argument_present = self.name in opts
        other_argument_present = self.not_required_if in opts
        if not argument_present and not other_argument_present:
            click.echo(
                click.style(
                    'At least one of the options: `{}` or `{}` is required\n'
                    .format(self.name, self.not_required_if),
                    fg='red') + ctx.get_help(),
                err=True)
            sys.exit(1)

        return super(NotRequiredIf, self).handle_parse_result(
            ctx, opts, args)


def output_user_friendly_logs(workflow_logs, steps):
    """Output workflow logs in a user-friendly manner.

    :param workflow_logs: Dictionary representing the workflow logs as they
        are returned from the REST API.
    :param steps: List of steps to show logs for.
    """
    # Colors matching the status colors of REANA-UI
    status_to_color_mapping = {
        'running': 'bright_blue',
        'finished': 'green',
        'failed': 'red',
    }
    key_to_description_mapping = {
        'workflow_uuid': 'Workflow ID',
        'compute_backend': 'Compute backend',
        'backend_job_id': 'Job ID',
        'docker_img': 'Docker image',
        'cmd': 'Command',
        'status': 'Status',
    }
    leading_mark = '==>'

    # REANA Workflow Engine logs
    if workflow_logs.get('workflow_logs'):
        click.secho('{} Workflow engine logs'.format(leading_mark),
                    bold=True, fg='yellow')
        click.echo(workflow_logs['workflow_logs'])

    if workflow_logs.get('engine_specific'):
        click.echo('\n')
        click.secho('{} Engine internal logs'.format(leading_mark),
                    fg='yellow')
        click.secho(workflow_logs['engine_specific'])

    returned_step_names = \
        set(workflow_logs['job_logs'][item]['job_name']
            for item in workflow_logs['job_logs'].keys())
    if steps:
        missing_steps = set(steps).difference(returned_step_names)
        if missing_steps:
            click.echo(
                click.style('The logs of step(s) {} were not found, '
                            'check for spelling mistakes in the step '
                            'names.'
                            .format(','.join(missing_steps)),
                            fg='red'))
    # Job logs
    if workflow_logs['job_logs']:
        click.echo('\n')
        click.secho('{} Job logs'.format(leading_mark), bold=True, fg='yellow')
    for job_id, logs_info in workflow_logs['job_logs'].items():
        if logs_info:
            job_name_or_id = logs_info['job_name'] or job_id
            click.secho('{0} Step: {1}'.format(
                leading_mark, job_name_or_id), bold=True,
                fg=status_to_color_mapping.get(logs_info['status']))
            logs_output = logs_info['logs']
            # extract already used fields
            del(logs_info['logs'])
            del(logs_info['job_name'])

            for key, value in logs_info.items():
                title = click.style('{mark} {description}:'.format(
                    mark=leading_mark,
                    description=key_to_description_mapping[key]),
                    fg=status_to_color_mapping.get(logs_info['status']))
                click.secho('{title} {value}'.format(title=title, value=value))
            # show actual log content
            if logs_output:
                click.secho('{mark} Logs:'.format(
                    mark=leading_mark),
                    fg=status_to_color_mapping.get(logs_info['status']))
                click.secho(logs_output)
            else:
                click.secho('{0} Step {1} emitted no logs.'
                            .format(leading_mark, job_name_or_id), bold=True)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client output related commands."""

import json
import logging
import os
import sys
import traceback

import click

from reana_client.api.utils import get_path_from_operation_id
from reana_client.cli.utils import (add_access_token_options,
                                    add_workflow_option, check_connection,
                                    filter_data, parse_parameters)
from reana_client.config import ERROR_MESSAGES, JSON, URL
from reana_client.errors import FileDeletionError, FileUploadError
from reana_client.utils import (get_workflow_root, load_reana_spec,
                                workflow_uuid_or_name)
from reana_commons.utils import click_table_printer


FILES_BLACKLIST = ('.git/', '/.git/')


@click.group(help='Workspace file management commands')
@click.pass_context
def files_group(ctx):
    """Top level wrapper for files related interactions."""
    logging.debug(ctx.info_name)


@files_group.command('ls')
@add_workflow_option
@check_connection
@click.option(
    '--format',
    '_filter',
    multiple=True,
    help='Format output according to column titles or column values. '
         'Use `<column_name>=<column_value>` format. For '
         'E.g. display FILES named data.txt '
         '`--format name=data.txt`.')
@click.option(
    '--json',
    'output_format',
    flag_value='json',
    default=None,
    help='Get output in JSON format.')
@click.option(
    '--url',
    'output_format',
    flag_value='url',
    default=None,
    help='Get URLs of output files.')
@add_access_token_options
@click.pass_context
def get_files(ctx, workflow, _filter,
              output_format, access_token):  # noqa: D301
    """List workspace files.

    The `ls` command lists workspace files of a workflow specified by the
    environment variable REANA_WORKON or provided as a command-line flag
    `--workflow` or `-w`.

    Examples: \n
    \t $ reana-client ls --workflow myanalysis.42
    """
    import tablib
    from reana_client.api.client import current_rs_api_client, list_files

    logging.debug('command: {}'.format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug('{param}: {value}'.format(param=p, value=ctx.params[p]))

    if _filter:
        parsed_filters = parse_parameters(_filter)
    if workflow:
        logging.info('Workflow "{}" selected'.format(workflow))
        try:
            response = list_files(workflow, access_token)
            headers = ['name', 'size', 'last-modified']
            data = []
            file_path = get_path_from_operation_id(
                current_rs_api_client.swagger_spec.spec_dict['paths'],
                'download_file')
            urls = []
            for file_ in response:
                if not file_['name'].startswith(FILES_BLACKLIST):
                    data.append(list(map(str,
                                         [file_['name'], file_['size'],
                                          file_['last-modified']])))
                    urls.append(ctx.obj.reana_server_url +
                                file_path.format(workflow_id_or_name=workflow,
                                                 file_name=file_['name']))
            tablib_data = tablib.Dataset()
            tablib_data.headers = headers
            for row in data:
                tablib_data.append(row)
            if output_format == URL:
                click.echo('\n'.join(urls))
            elif _filter:
                tablib_data, filtered_headers = \
                    filter_data(parsed_filters, headers, tablib_data)
                if output_format == JSON:
                    click.echo(json.dumps(tablib_data))
                else:
                    tablib_data = [list(item.values()) for item in tablib_data]
                    click_table_printer(filtered_headers, filtered_headers,
                                        tablib_data)
            else:
                if output_format == JSON:
                    click.echo(tablib_data.export(output_format))
                else:
                    click_table_printer(headers, _filter, data)

        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))

            click.echo(
                click.style('Something went wrong while retrieving file list'
                            ' for workflow {0}:\n{1}'.format(workflow,
                                                             str(e)),
                            fg='red'),
                err=True)


@files_group.command('download')
@click.argument(
    'filenames',
    metavar='FILES',
    nargs=-1)
@add_workflow_option
@check_connection
@click.option(
    '-o',
    '--output-directory',
    default=os.getcwd(),
    help='Path to the directory where files will be downloaded.')
@add_access_token_options
@click.pass_context
def download_files(ctx, workflow, filenames,
                   output_directory, access_token):  # noqa: D301
    """Download workspace files.

    The `download` command allows to download workspace files. By default, the
    files specified in the workflow specification as outputs are downloaded.
    You can also specify the individual files you would like to download, see
    examples below. Note that downloading directories is not yet supported.

    Examples: \n
    \t $ reana-client download # download all output files \n
    \t $ reana-client download mydata.tmp outputs/myplot.png
    """
    from reana_client.api.client import download_file

    logging.debug('command: {}'.format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug('{param}: {value}'.format(param=p, value=ctx.params[p]))

    if not filenames:
        reana_spec = load_reana_spec(os.path.join(get_workflow_root(),
                                     'reana.yaml'),
                                     False)
        if 'outputs' in reana_spec:
            filenames = reana_spec['outputs'].get('files') or []

    if workflow:
        for file_name in filenames:
            try:
                binary_file = download_file(workflow,
                                            file_name,
                                            access_token)

                logging.info('{0} binary file downloaded ... writing to {1}'.
                             format(file_name, output_directory))

                outputs_file_path = os.path.join(output_directory, file_name)
                if not os.path.exists(os.path.dirname(outputs_file_path)):
                    os.makedirs(os.path.dirname(outputs_file_path))

                with open(outputs_file_path, 'wb') as f:
                    f.write(binary_file)
                click.secho(
                    'File {0} downloaded to {1}.'.format(
                        file_name, output_directory),
                    fg='green')
            except OSError as e:
                logging.debug(traceback.format_exc())
                logging.debug(str(e))
                click.echo(
                    click.style('File {0} could not be written.'.
                                format(file_name),
                                fg='red'), err=True)
            except Exception as e:
                logging.debug(traceback.format_exc())
                logging.debug(str(e))
                click.echo(click.style('File {0} could not be downloaded: {1}'.
                                       format(file_name, e), fg='red'),
                           err=True)


@files_group.command('upload')
@click.argument(
    'filenames',
    metavar='SOURCES',
    type=click.Path(exists=True, resolve_path=True),
    nargs=-1)
@add_workflow_option
@check_connection
@add_access_token_options
@click.pass_context
def upload_files(ctx, workflow, filenames, access_token):  # noqa: D301
    """Upload files and directories to workspace.

    The `upload` command allows to upload workflow input files and
    directories. The SOURCES argument can be repeated and specifies which files
    and directories are to be uploaded, see examples below. The default
    behaviour is to upload all input files and directories specified in the
    reana.yaml file.

    Examples: \n
    \t $ reana-client upload -w myanalysis.42 \n
    \t $ reana-client upload -w myanalysis.42 code/mycode.py
    """
    from reana_client.api.client import upload_to_server

    logging.debug('command: {}'.format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug('{param}: {value}'.format(param=p, value=ctx.params[p]))

    if not filenames:
        reana_spec = load_reana_spec(os.path.join(get_workflow_root(),
                                     'reana.yaml'),
                                     False)
        if 'inputs' in reana_spec:
            filenames = []
            filenames += [os.path.join(get_workflow_root(), f)
                          for f in reana_spec['inputs'].get('files') or []]
            filenames += [os.path.join(get_workflow_root(), d)
                          for d in reana_spec['inputs'].
                          get('directories') or []]

    if workflow:
        if filenames:
            for filename in filenames:
                try:
                    response = upload_to_server(workflow,
                                                filename,
                                                access_token)
                    for file_ in response:
                        if file_.startswith('symlink:'):
                            click.echo(
                                click.style('Symlink resolved to {}. Uploaded'
                                            ' hard copy.'.
                                            format(file_[len('symlink:'):]),
                                            fg='green'))
                        else:
                            click.echo(
                                click.style('File {} was successfully '
                                            'uploaded.'.format(file_),
                                            fg='green'))
                except FileNotFoundError as e:
                    logging.debug(traceback.format_exc())
                    logging.debug(str(e))
                    click.echo(
                        click.style(
                            'File {0} could not be uploaded: {0} does not'
                            ' exist.'.format(filename),
                            fg='red'),
                        err=True)
                    if 'invoked_by_subcommand' in ctx.parent.__dict__:
                        sys.exit(1)
                except FileUploadError as e:
                    logging.debug(traceback.format_exc())
                    logging.debug(str(e))
                    click.echo(
                        click.style(
                            'Something went wrong while uploading {0}.\n{1}'.
                            format(filename, str(e)),
                            fg='red'),
                        err=True)
                    if 'invoked_by_subcommand' in ctx.parent.__dict__:
                        sys.exit(1)
                except Exception as e:
                    logging.debug(traceback.format_exc())
                    logging.debug(str(e))
                    click.echo(
                        click.style(
                            'Something went wrong while uploading {}'.
                            format(filename),
                            fg='red'),
                        err=True)
                    if 'invoked_by_subcommand' in ctx.parent.__dict__:
                        sys.exit(1)


@files_group.command('rm')
@click.argument(
    'filenames',
    metavar='SOURCES',
    nargs=-1)
@add_workflow_option
@check_connection
@add_access_token_options
@click.pass_context
def delete_files(ctx, workflow, filenames, access_token):  # noqa: D301
    """Delete files from workspace.

    The `rm` command allow to delete files and directories from workspace.
    Note that you can use glob to remove similar files.

    Examples:\n
    \t $ reana-client rm -w myanalysis.42 data/mydata.csv \n
    \t $ reana-client rm -w myanalysis.42 'code/\*'
    """
    from reana_client.api.client import delete_file

    logging.debug('command: {}'.format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug('{param}: {value}'.format(param=p, value=ctx.params[p]))

    if workflow:
        for filename in filenames:
            try:
                response = delete_file(workflow, filename, access_token)
                freed_space = 0
                for file_ in response['deleted']:
                    freed_space += response['deleted'][file_]['size']
                    click.echo(click.style(
                        'File {} was successfully deleted.'.
                        format(file_), fg='green'))
                for file_ in response['failed']:
                    click.echo(
                        click.style(
                            'Something went wrong while deleting {}.\n{}'.
                            format(file_, response['failed'][file_]['error']),
                            fg='red'), err=True)
                if freed_space:
                    click.echo(click.style('{} bytes freed up.'.format(
                        freed_space), fg='green'))
            except FileDeletionError as e:
                click.echo(click.style(str(e), fg='red'), err=True)
                if 'invoked_by_subcommand' in ctx.parent.__dict__:
                    sys.exit(1)
            except Exception as e:
                logging.debug(traceback.format_exc())
                logging.debug(str(e))
                click.echo(
                    click.style(
                        'Something went wrong while deleting {}'.
                        format(filename),
                        fg='red'),
                    err=True)
                if 'invoked_by_subcommand' in ctx.parent.__dict__:
                    sys.exit(1)


@files_group.command('mv')
@click.argument('source')
@click.argument('target')
@add_workflow_option
@check_connection
@add_access_token_options
@click.pass_context
def move_files(ctx, source, target, workflow, access_token):  # noqa: D301
    """Move files within workspace.

    The `mv` command allow to move the files within workspace.

    Examples:\n
    \t $ reana-client mv data/input.txt input/input.txt
    """
    from reana_client.api.client import (get_workflow_status, list_files,
                                         mv_files)

    logging.debug('command: {}'.format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug('{param}: {value}'.format(param=p, value=ctx.params[p]))

    if workflow:
        try:
            current_status = get_workflow_status(workflow,
                                                 access_token).get('status')
            if current_status == 'running':
                click.echo(
                    click.style('File(s) could not be moved for running '
                                'workflow', fg='red'),
                    err=True)
                sys.exit(1)
            files = list_files(workflow, access_token)
            current_files = [file['name'] for file in files]
            if not any(source in item for item in current_files):
                click.echo(
                    click.style('Source file(s) {} does not exist in '
                                'workspace {}'.format(source,
                                                      current_files),
                                fg='red'),
                    err=True)
                sys.exit(1)
            response = mv_files(source, target, workflow, access_token)
            click.echo(click.style(
                '{} was successfully moved to {}.'.
                format(source, target), fg='green'))
        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            click.echo(
                click.style('Something went wrong. {}'.format(e), fg='red'),
                err=True)


@files_group.command('du')
@add_workflow_option
@check_connection
@add_access_token_options
@click.option(
    '-s',
    '--summarize',
    is_flag=True,
    help='Display total.')
@click.option(
    '-b',
    '--bytes',
    'block_size',
    flag_value='b',
    help='Print size in bytes.')
@click.option(
    '-k',
    '--kilobytes',
    'block_size',
    flag_value='k',
    help='Print size in kilobytes.')
@click.pass_context
def workflow_disk_usage(ctx, workflow, access_token, summarize, block_size):  # noqa: D301
    """Get workspace disk usage.

    The `du` command allows to chech the disk usage of given workspace.

    Examples: \n
    \t $ reana-client du -w myanalysis.42 -s \n
    \t $ reana-client du -w myanalysis.42 --bytes
    """
    from reana_client.api.client import get_workflow_disk_usage

    logging.debug('command: {}'.format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug('{param}: {value}'.format(param=p, value=ctx.params[p]))

    if workflow:
        try:
            parameters = {'summarize': summarize, 'block_size': block_size}
            response = get_workflow_disk_usage(workflow,
                                               parameters,
                                               access_token)
            headers = ['size', 'name']
            data = []
            for disk_usage_info in response['disk_usage_info']:
                if not disk_usage_info['name'].startswith(FILES_BLACKLIST):
                    data.append([disk_usage_info['size'],
                                '.{}'.format(disk_usage_info['name'])])
            click_table_printer(headers, [], data)
        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            click.echo(
                click.style('Disk usage could not be retrieved: \n{}'
                            .format(str(e)), fg='red'),
                err=True)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA REST API client."""

import json
import logging
import os
import traceback
from functools import partial

import requests
from bravado.exception import HTTPError
from reana_client.config import ERROR_MESSAGES, WORKFLOW_ENGINES
from reana_client.errors import FileDeletionError, FileUploadError
from reana_client.utils import (_validate_reana_yaml, get_workflow_root,
                                is_uuid_v4)
from reana_commons.api_client import get_current_api_client
from reana_commons.errors import (REANASecretAlreadyExists,
                                  REANASecretDoesNotExist)
from werkzeug.local import LocalProxy

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

current_rs_api_client = LocalProxy(
    partial(get_current_api_client, component='reana-server'))


def ping():
    """Health check REANA server."""
    try:
        response, http_response = current_rs_api_client.api.ping().result()
        if http_response.status_code == 200:
            return response['message']
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'REANA server health check failed: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def get_workflows(access_token, type, verbose=False, block_size=None):
    """List all existing workflows."""
    try:
        response, http_response = current_rs_api_client.api.\
            get_workflows(access_token=access_token, verbose=verbose,
                          type=type, block_size=block_size).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'The list of workflows could not be retrieved: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def get_workflow_status(workflow, access_token):
    """Get status of previously created workflow."""
    try:
        response, http_response = current_rs_api_client.api\
            .get_workflow_status(
                workflow_id_or_name=workflow,
                access_token=access_token).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Analysis status could not be retrieved: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def create_workflow(reana_specification, name, access_token):
    """Create a workflow."""
    try:
        (response,
            http_response) = current_rs_api_client.api.create_workflow(
                reana_specification=json.loads(json.dumps(
                    reana_specification, sort_keys=True)),
                workflow_name=name,
                access_token=access_token).result()
        if http_response.status_code == 201:
            return response
        else:
            raise Exception(
                "Expected status code 201 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Workflow creation failed: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def create_workflow_from_json(workflow_json, name, access_token,
                              parameters=None, workflow_engine='yadage',
                              outputs=None):
    """Create a workflow from json specification.

    :param workflow_json: workflow specification in json format.
    :param name: name or UUID of the workflow to be started.
    :param access_token: Access token of the current user.
    :param parameters: workflow input parameters dictionary.
    :param workflow_engine: one of the workflow engines (yadage, serial, cwl)
    :param outputs: dictionary with expected workflow outputs.

    :Example:

      .. code:: python

        create_workflow_from_json(
            workflow_json=workflow_json,
            name='workflow_name.1',
            access_token='access_token',
            parameters={'files': ['file.txt'], 'parameters': {'key': 'value'}},
            workflow_engine='serial')
    """
    if is_uuid_v4(name):
        raise ValueError('Workflow name cannot be a valid UUIDv4')
    if not access_token:
        raise Exception(ERROR_MESSAGES['missing_access_token'])
    if os.environ.get('REANA_SERVER_URL') is None:
        raise Exception('Environment variable REANA_SERVER_URL is not set')
    workflow_engine = workflow_engine.lower()
    if workflow_engine not in WORKFLOW_ENGINES:
        raise Exception('Workflow engine - {} not found. You must use one of '
                        'these engines - {}'.format(workflow_engine,
                                                    WORKFLOW_ENGINES))
    try:
        reana_yaml = {}
        reana_yaml['workflow'] = {'specification': workflow_json}
        reana_yaml['workflow']['type'] = workflow_engine
        if parameters:
            reana_yaml['inputs'] = parameters
        if outputs:
            reana_yaml['outputs'] = outputs
        _validate_reana_yaml(reana_yaml)
        reana_specification = reana_yaml
        (response,
            http_response) = current_rs_api_client.api.create_workflow(
                reana_specification=json.loads(json.dumps(
                    reana_specification, sort_keys=True)),
                workflow_name=name,
                access_token=access_token).result()
        if http_response.status_code == 201:
            return response
        else:
            raise Exception(
                "Expected status code 201 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Workflow creation failed: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def start_workflow(workflow, access_token, parameters):
    """Start a workflow.

    :param workflow: name or id of previously created workflow.
    :param access_token: access token of the current user.
    :param parameters: dict of workflow parameters to override the original
        ones (after workflow creation).
    :param restart: boolean if workflow should be restarted on the same
        workspace.
    """
    try:
        (response,
         http_response) = current_rs_api_client.api.start_workflow(
            workflow_id_or_name=workflow,
            access_token=access_token,
            parameters=parameters).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Workflow run could not be started: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def upload_file(workflow_id, file_, file_name, access_token):
    """Upload file to workflow workspace.

    :param workflow_id: UID which identifies the workflow.
    :param file_: content of a file that will be uploaded.
    :param file_name: name of a file that will be uploaded.
    :param access_token: access token of the current user.
    """
    from reana_client.utils import get_api_url
    try:
        endpoint = \
            current_rs_api_client.api.upload_file.operation.path_name.format(
                workflow_id_or_name=workflow_id)
        http_response = requests.post(urljoin(get_api_url(), endpoint),
                                      data=file_,
                                      params={'file_name': file_name,
                                              'access_token': access_token},
                                      headers={'Content-Type':
                                               'application/octet-stream'},
                                      verify=False)
        return http_response.json()
    except requests.exceptions.ConnectionError:
        logging.debug(
            'File could not be uploaded.', exc_info=True)
        raise Exception(
            'Could not connect to the server {}'.format(get_api_url()))
    except requests.exceptions.HTTPError as e:
        logging.debug(
            'The server responded with an HTTP error code.', exc_info=True)
        raise Exception(
            'Unexpected response from the server: \n{}'.format(e.response))
    except requests.exceptions.Timeout:
        logging.debug(
            'Timeout while trying to establish connection.', exc_info=True)
        raise Exception('The request to the server has timed out.')
    except requests.exceptions.RequestException:
        logging.debug(
            'Something went wrong while connecting to the server.',
            exc_info=True)
        raise Exception('The request to the server has failed for an '
                        'unknown reason.')
    except Exception as e:
        raise e


def get_workflow_logs(workflow_id, access_token, steps=None):
    """Get logs from a workflow engine."""
    try:
        (response,
            http_response) = current_rs_api_client.api.get_workflow_logs(
                workflow_id_or_name=workflow_id,
                steps=steps,
                access_token=access_token).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Workflow logs could not be retrieved: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def download_file(workflow_id, file_name, access_token):
    """Download the requested file if it exists.

    :param workflow_id: UUID which identifies the workflow.
    :param file_name: File name or path to the file requested.
    :returns: .
    """
    try:
        from reana_client.utils import get_api_url
        logging.getLogger("urllib3").setLevel(logging.CRITICAL)
        endpoint = \
            current_rs_api_client.api.download_file.operation.path_name.format(
                workflow_id_or_name=workflow_id, file_name=file_name)
        http_response = requests.get(urljoin(get_api_url(), endpoint),
                                     params={'file_name': file_name,
                                             'access_token': access_token},
                                     verify=False)
        if http_response.status_code == 200:
            return http_response.content
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Output file could not be downloaded: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def delete_file(workflow_id, file_name, access_token):
    """Delete the requested file if it exists.

    :param workflow_id: UUID which identifies the workflow.
    :param file_name: File name or path to the file requested.
    """
    try:
        (response,
            http_response) = current_rs_api_client.api.delete_file(
                workflow_id_or_name=workflow_id,
                file_name=file_name,
                access_token=access_token).result()
        if http_response.status_code == 200 and (response['deleted'] or
                                                 response['failed']):
            return response
        elif not (response['deleted'] or response['failed']):
            raise FileDeletionError('{} did not match any existing '
                                    'file.'.format(file_name))
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'File could not be downloaded: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def list_files(workflow_id, access_token):
    """Return the list of file for a given workflow workspace.

    :param workflow_id: UUID which identifies the workflow.
    :returns: A list of dictionaries composed by the `name`, `size` and
                `last-modified`.
    """
    try:
        (response,
            http_response) = current_rs_api_client.api.get_files(
                workflow_id_or_name=workflow_id,
                access_token=access_token).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'File list could not be retrieved: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def upload_to_server(workflow, paths, access_token):
    """Upload file or directory to REANA server.

    Shared e.g. by `code upload` and `inputs upload`.

    :param workflow: ID of that workflow whose workspace should be
        used to store the files.
    :param paths: Absolute filepath(s) of files to be uploaded.
    """
    if not workflow:
        raise ValueError(
            'Workflow name or id must be provided')
    if not paths:
        raise ValueError(
            'Please provide path(s) to file(s) that '
            'should be uploaded to workspace.')

    logging.info('Workflow "{}" selected'.format(workflow))

    # Check if multiple paths were given and iterate over them
    if type(paths) is list or type(paths) is tuple:
        for path in paths:
            upload_to_server(workflow, path, access_token)
    # `paths` points to a single file or directory
    else:
        path = paths
        if '..' in paths.split('/'):
            raise FileUploadError('Path cannot contain ".."')

        # Check if input is a directory and upload everything
        # including subdirectories.

        if os.path.isdir(path):
            logging.debug("'{}' is a directory.".format(path))
            logging.info("Uploading contents of folder '{}' ...".format(path))
            for root, dirs, files in os.walk(path, topdown=False):
                uploaded_files = []
                for next_path in files + dirs:
                    next_uploaded_files = upload_to_server(
                        workflow, os.path.join(root, next_path),
                        access_token)
                    if next_uploaded_files:
                        uploaded_files.extend(next_uploaded_files)
            return uploaded_files

        # Check if input is an absolute path and upload file.
        else:
            symlink = False
            if os.path.islink(path):
                path = os.path.realpath(path)
                logging.info(
                    'Symlink {} found, uploading'
                    ' hard copy.'.
                    format(path))
                symlink = True
            with open(path, 'rb') as f:
                fname = os.path.basename(f.name)
                workflow_root = get_workflow_root()
                if not path.startswith(workflow_root):
                    raise FileUploadError(
                        'Files and directories to be uploaded '
                        'must be under the workflow root directory.')
                # Calculate the path that will store the file
                # in the workflow controller, by subtracting
                # the workflow root path from the file path
                save_path = path.replace(workflow_root, '')
                # Remove prepending dirs named "." or as the upload type
                while len(save_path.split('/')) > 1 and \
                        save_path.split('/')[0] == '.':
                    save_path = "/".join(
                        save_path.strip("/").split('/')[1:])
                logging.debug("'{}' is an absolute filepath."
                              .format(os.path.basename(fname)))
                logging.info("Uploading '{}' ...".format(fname))
                try:
                    upload_file(workflow, f, save_path, access_token)
                    logging.info("File '{}' was successfully "
                                 "uploaded.".format(fname))
                    if symlink:
                        save_path = 'symlink:{}'.format(save_path)
                    return [save_path]
                except Exception as e:
                    logging.debug(traceback.format_exc())
                    logging.debug(str(e))
                    logging.info("Something went wrong while uploading {}"
                                 .format(fname))


def get_workflow_parameters(workflow, access_token):
    """Get parameters of previously created workflow."""
    try:
        response, http_response = current_rs_api_client.api\
            .get_workflow_parameters(
                workflow_id_or_name=workflow,
                access_token=access_token)\
            .result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Workflow parameters could not be retrieved: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def delete_workflow(workflow, all_runs, hard_delete,
                    workspace, access_token):
    """Delete a workflow."""
    try:
        parameters = {
            'all_runs': True if all_runs == 1 else False,
            'hard_delete': True if hard_delete == 1 else False,
            'workspace': True if hard_delete == 1 or workspace == 1 else False
        }
        (response,
            http_response) = current_rs_api_client.api.set_workflow_status(
            workflow_id_or_name=workflow,
            status='deleted',
            access_token=access_token,
            parameters=parameters).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Workflow run could not be deleted: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def stop_workflow(workflow, force_stop, access_token):
    """Stop a workflow."""
    try:
        parameters = {'force_stop': force_stop}
        (response, http_response) = current_rs_api_client.api\
            .set_workflow_status(
            workflow_id_or_name=workflow,
            status='stop',
            access_token=access_token,
            parameters=parameters).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))
    except HTTPError as e:
        logging.debug(
            'Workflow run could not be stopped: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def diff_workflows(workflow_id_a, workflow_id_b,
                   brief, access_token, context_lines):
    """Return the list of differences between two workflows.

    :param workflow_id_a: UUID which identifies the first workflow.
    :param workflow_id_b: UUID which identifies the second workflow.
    :param brief: Flag specifying desired detail in diff.
    :param context_lines: Optional parameter to set the number of
                          context lines shown in the diff output.
    :param access_token: API token of user requesting diff.
    :returns: A list of dictionaries composed by `asset`, `type`, `lines`,
        `a` and `b`. Asset refers to the workflow asset where a
        difference was found, type refers to the asset type, lines refer
        to the lines of the file where the differences are and a, b fields
        are the actual lines that differ.
    """
    try:
        (response,
         http_response) = current_rs_api_client.api.get_workflow_diff(
             workflow_id_or_name_a=workflow_id_a,
             workflow_id_or_name_b=workflow_id_b,
             brief=brief,
             context_lines=context_lines,
             access_token=access_token).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'File list could not be retrieved: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def open_interactive_session(workflow, access_token,
                             interactive_session_type,
                             interactive_session_configuration):
    """Open an interactive session inside the workflow workspace.

    :param workflow: Workflow which workspace will be available inside the
        interactive session.
    :param access_token: Workflow owner REANA access token.
    :param interactive_session_type: Type of interactive session to spawn.
    :param interactive_session_configuration: Specific configuration for
        the interactive session.

    :return: Gives the relative path to the interactive service.
    """
    try:
        (response, http_response) = current_rs_api_client.api\
            .open_interactive_session(
            workflow_id_or_name=workflow,
            access_token=access_token,
            interactive_session_type=interactive_session_type,
            interactive_session_configuration=interactive_session_configuration
        ).result()
        if http_response.status_code == 200:
            return response['path']
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))
    except HTTPError as e:
        logging.debug(
            'Interactive session could not be opened: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def close_interactive_session(workflow, access_token):
    """Close an interactive workflow session.

    :param workflow: Workflow name to close.
    :param access_token: Workflow owner REANA access token.

    :return: Gives the relative path to the interactive service.
    """
    try:
        (response, http_response) = current_rs_api_client.api\
            .close_interactive_session(
            workflow_id_or_name=workflow,
            access_token=access_token,
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))
    except HTTPError as e:
        logging.debug(
            'Interactive session could not be closed: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def mv_files(source, target, workflow, access_token):
    """Move target file(s) within workspace.

    :param source: source filename or path.
    :param target: target filename or path.
    :param workflow_id: UUID which identifies the workflowself.
    :param access_token: token of user.
    """
    try:
        (response, http_response) = current_rs_api_client.api.move_files(
            source=source,
            target=target,
            workflow_id_or_name=workflow,
            access_token=access_token).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Files move command failed: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def get_workflow_disk_usage(workflow, parameters, access_token):
    """Display disk usage workflow."""
    try:
        (response, http_response) = current_rs_api_client.api\
            .get_workflow_disk_usage(
            workflow_id_or_name=workflow,
            parameters=parameters,
            access_token=access_token).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Workflow disk usage could not be retrieved: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def add_secrets(secrets, overwrite, access_token):
    """Add new secrets.

    :param secrets: dictionary containing all the secrets to be sent.
      The dictionary with secret names for keys and for each key there is
       a dictionary with two fields:
      - 'value':  a base64 encoded file or literal string
      - 'type': 'file' or 'env'
    :param overwrite: whether secrets should be overwritten when they
     already exist.
    :param access_token: access token of the current user.

    """
    try:
        (response,
            http_response) = current_rs_api_client.api.add_secrets(
            secrets=secrets,
            access_token=access_token,
            overwrite=overwrite).result()
        if http_response.status_code == 201:
            return response
        else:
            raise Exception(
                "Expected status code 201 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Secrets could not be added: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        if e.status_code == 409:
            raise REANASecretAlreadyExists()
        else:
            raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def delete_secrets(secrets, access_token):
    """Delete a list of secrets.

    :param secrets: List of secret names to be deleted.
    :param access_token: access token of the current user.

    """
    try:
        (response,
            http_response) = current_rs_api_client.api.delete_secrets(
            secrets=secrets,
            access_token=access_token).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        if e.response.status_code == 404:
            raise REANASecretDoesNotExist(e.response.json())
        else:
            logging.debug(
                'Secrets could not be deleted: '
                '\nStatus: {}\nReason: {}\n'
                'Message: {}'.format(e.response.status_code,
                                     e.response.reason,
                                     e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e


def list_secrets(access_token):
    """List user secrets.

    :param access_token: access token of the current user.

    """
    try:
        (response,
            http_response) = current_rs_api_client.api.get_secrets(
            access_token=access_token).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(
                    status_code=http_response.status_code))

    except HTTPError as e:
        logging.debug(
            'Secrets could not be listed: '
            '\nStatus: {}\nReason: {}\n'
            'Message: {}'.format(e.response.status_code,
                                 e.response.reason,
                                 e.response.json()['message']))
        raise Exception(e.response.json()['message'])
    except Exception as e:
        raise e



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2019 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client"""

from __future__ import absolute_import, print_function

import os
import re

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    'pytest-reana==0.7.0.dev20191219',
]

extras_require = {
    'docs': [
        'Sphinx>=1.5.1',
        'sphinx-rtd-theme>=0.1.9',
        'sphinx-click>=1.0.4',
    ],
    'tests': tests_require,
}

extras_require['all'] = []
for key, reqs in extras_require.items():
    if ':' == key[0]:
        continue
    extras_require['all'].extend(reqs)

setup_requires = [
    'pytest-runner>=2.7',
]

install_requires = [
    'PyYAML>=5.1',
    'typing>=3.7.4 ; python_version=="2.7"',  # workaround for CWL deps
    'click>=7',
    'cryptography>=2.7',
    'cwltool==1.0.20191022103248',
    'pyOpenSSL>=19.0.0',  # FIXME remove once yadage-schemas solves deps.
    'jsonpointer>=2.0',
    'reana-commons>=0.7.0.dev20200217,<0.8.0',
    'rfc3987>=1.3.8',  # FIXME remove once yadage-schemas solves deps.
    'six==1.12.0',
    'strict-rfc3339>=0.7',  # FIXME remove once yadage-schemas solves deps.
    'tablib>=0.12.1,<0.13',
    'webcolors==1.9.1',  # FIXME remove once yadage-schemas solves deps.
    'werkzeug>=0.14.1',
    'yadage-schemas==0.10.6',
]

packages = find_packages()


# Get the version string. Cannot be done with import!
with open(os.path.join('reana_client', 'version.py'), 'rt') as f:
    version = re.search(
        '__version__\s*=\s*"(?P<version>.*)"\n',
        f.read()
    ).group('version')

setup(
    name='reana-client',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    author='REANA',
    author_email='info@reana.io',
    url='https://github.com/reanahub/reana-client',
    packages=['reana_client', ],
    zip_safe=False,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'reana-client = reana_client.cli:cli',
            'reana-cwl-runner = reana_client.cli.cwl_runner:cwl_runner'
        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client utils."""
import base64
import json
import logging
import os
import subprocess
import sys
import traceback
from uuid import UUID

import click
import yadageschemas
import yaml
from jsonschema import ValidationError, validate
from reana_commons.serial import serial_load
from reana_commons.utils import get_workflow_status_change_verb

from reana_client.config import reana_yaml_schema_file_path


def workflow_uuid_or_name(ctx, param, value):
    """Get UUID of workflow from configuration / cache file based on name."""
    if not value:
        click.echo(click.style(
            'Workflow name must be provided either with '
            '`--workflow` option or with REANA_WORKON '
            'environment variable', fg='red'),
            err=True)
    else:
        return value


def yadage_load(workflow_file, toplevel='.'):
    """Validate and return yadage workflow specification.

    :param workflow_file: A specification file compliant with
        `yadage` workflow specification.
    :type workflow_file: string
    :param toplevel: URL/path for the workflow file
    :type toplevel: string

    :returns: A dictionary which represents the valid `yadage` workflow.
    """
    schema_name = 'yadage/workflow-schema'
    schemadir = None

    specopts = {
        'toplevel': toplevel,
        'schema_name': schema_name,
        'schemadir': schemadir,
        'load_as_ref': False,
    }

    validopts = {
        'schema_name': schema_name,
        'schemadir': schemadir,
    }

    return yadageschemas.load(spec=workflow_file, specopts=specopts,
                              validopts=validopts, validate=True)


def cwl_load(workflow_file):
    """Validate and return cwl workflow specification.

    :param workflow_file: A specification file compliant with
        `cwl` workflow specification.
    :returns: A dictionary which represents the valid `cwl` workflow.
    """
    result = \
        subprocess.run(
            ['cwltool', '--pack', '--quiet', workflow_file],
            stdout=subprocess.PIPE)
    value = result.stdout.decode('utf-8')
    return json.loads(value)


workflow_load = {
    'yadage': yadage_load,
    'cwl': cwl_load,
    'serial': serial_load,
}
"""Dictionary to extend with new workflow specification loaders."""


def load_workflow_spec(workflow_type, workflow_file, **kwargs):
    """Validate and return machine readable workflow specifications.

    :param workflow_type: A supported workflow specification type.
    :param workflow_file: A workflow file compliant with `workflow_type`
        specification.
    :returns: A dictionary which represents the valid workflow specification.
    """
    return workflow_load[workflow_type](workflow_file, **kwargs)


def load_reana_spec(filepath, skip_validation=False):
    """Load and validate reana specification file.

    :raises IOError: Error while reading REANA spec file from given filepath`.
    :raises ValidationError: Given REANA spec file does not validate against
        REANA specification.
    """
    try:
        with open(filepath) as f:
            reana_yaml = yaml.load(f.read(), Loader=yaml.FullLoader)

        if not skip_validation:
            logging.info('Validating REANA specification file: {filepath}'
                         .format(filepath=filepath))
            _validate_reana_yaml(reana_yaml)

        kwargs = {}
        if reana_yaml['workflow']['type'] == 'serial':
            kwargs['specification'] = reana_yaml['workflow'].\
                get('specification')
            kwargs['parameters'] = \
                reana_yaml.get('inputs', {}).get('parameters', {})
            kwargs['original'] = True

        reana_yaml['workflow']['specification'] = load_workflow_spec(
            reana_yaml['workflow']['type'],
            reana_yaml['workflow'].get('file'),
            **kwargs
        )

        if reana_yaml['workflow']['type'] == 'cwl' and \
                'inputs' in reana_yaml:
            with open(reana_yaml['inputs']['parameters']['input']) as f:
                reana_yaml['inputs']['parameters'] = \
                    yaml.load(f, Loader=yaml.FullLoader)

        return reana_yaml
    except IOError as e:
        logging.info(
            'Something went wrong when reading specifications file from '
            '{filepath} : \n'
            '{error}'.format(filepath=filepath, error=e.strerror))
        raise e
    except Exception as e:
        raise e


def _validate_reana_yaml(reana_yaml):
    """Validate REANA specification file according to jsonschema.

    :param reana_yaml: Dictionary which represents REANA specifications file.
    :raises ValidationError: Given REANA spec file does not validate against
        REANA specification schema.
    """
    try:
        with open(reana_yaml_schema_file_path, 'r') as f:
            reana_yaml_schema = json.loads(f.read())

            validate(reana_yaml, reana_yaml_schema)

    except IOError as e:
        logging.info(
            'Something went wrong when reading REANA validation schema from '
            '{filepath} : \n'
            '{error}'.format(filepath=reana_yaml_schema_file_path,
                             error=e.strerror))
        raise e
    except ValidationError as e:
        logging.info('Invalid REANA specification: {error}'
                     .format(error=e.message))
        raise e


def is_uuid_v4(uuid_or_name):
    """Check if given string is a valid UUIDv4."""
    # Based on https://gist.github.com/ShawnMilo/7777304
    try:
        uuid = UUID(uuid_or_name, version=4)
    except Exception:
        return False

    return uuid.hex == uuid_or_name.replace('-', '')


def get_workflow_name_and_run_number(workflow_name):
    """Return name and run_number of a workflow.

    :param workflow_name: String representing Workflow name.

        Name might be in format 'reana.workflow.123' with arbitrary
        number of dot-delimited substrings, where last substring specifies
        the run number of the workflow this workflow name refers to.

        If name does not contain a valid run number, name without run number
        is returned.
    """
    # Try to split a dot-separated string.
    try:
        name, run_number = workflow_name.split('.', 1)

        try:
            float(run_number)
        except ValueError:
            # `workflow_name` was split, so it is a dot-separated string
            # but it didn't contain a valid `run_number`.
            # Assume that this dot-separated string is the name of
            # the workflow and return just this without a `run_number`.
            return workflow_name, ''

        return name, run_number

    except ValueError:
        # Couldn't split. Probably not a dot-separated string.
        # Return the name given as parameter without a `run_number`.
        return workflow_name, ''


def get_workflow_root():
    """Return the current workflow root directory."""
    reana_yaml = 'reana.yaml'
    workflow_root = os.getcwd()

    while True:
        file_list = os.listdir(workflow_root)
        parent_dir = os.path.dirname(workflow_root)
        if reana_yaml in file_list:
            break
        else:
            if workflow_root == parent_dir:
                click.echo(click.style(
                    'Not a workflow directory (or any of the parent'
                    ' directories).\nPlease upload from inside'
                    ' the directory containing the reana.yaml '
                    'file of your workflow.', fg='red'))
                sys.exit(1)
            else:
                workflow_root = parent_dir
    workflow_root += '/'
    return workflow_root


def validate_cwl_operational_options(operational_options):
    """Validate cwl operational options."""
    forbidden_args = ['--debug', '--tmpdir-prefix', '--tmp-outdir-prefix'
                      '--default-container', '--outdir']
    for option in operational_options:
        if option in forbidden_args:
            click.echo('Operational option {0} are not allowed. \n'
                       .format(operational_options), err=True)
            sys.exit(1)
    cmd = 'cwltool --version {0}'.format(' '.join(operational_options))
    try:
        subprocess.check_call(cmd, shell=True)
    except subprocess.CalledProcessError as err:
        click.echo('Operational options {0} are not valid. \n'
                   '{1}'.format(operational_options, err), err=True)
        sys.exit(1)


def validate_serial_operational_options(operational_options):
    """Return validated serial operational options."""
    try:
        return dict(p.split('=') for p in operational_options)
    except Exception as err:
        click.echo('Operational options {0} are not valid. \n'
                   '{1}'.format(operational_options, err), err=True)
        sys.exit(1)


def validate_input_parameters(live_parameters, original_parameters):
    """Return validated input parameters."""
    parsed_input_parameters = dict(live_parameters)
    for parameter in parsed_input_parameters.keys():
        if parameter not in original_parameters:
            click.echo(
                click.style('Given parameter - {0}, is not in '
                            'reana.yaml'.format(parameter),
                            fg='red'),
                err=True)
            del live_parameters[parameter]
    return live_parameters


def get_workflow_status_change_msg(workflow, status):
    """Choose verb conjugation depending on status.

    :param workflow: Workflow name whose status changed.
    :param status: String which represents the status the workflow changed to.
    """
    return '{workflow} {verb} {status}'.format(
        workflow=workflow, verb=get_workflow_status_change_verb(status),
        status=status)


def parse_secret_from_literal(literal):
    """Parse a literal string, into a secret dict.

    :param literal: String containg a key and a value. (e.g. 'KEY=VALUE')
    :returns secret: Dictionary in the format suitable for sending
    via http request.
    """
    try:
        key, value = literal.split('=', 1)
        secret = {
            key: {
                'value': base64.b64encode(
                    value.encode('utf-8')
                ).decode('utf-8'),
                'type': 'env'
            }
        }
        return secret

    except ValueError as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        click.echo(
            click.style(
                'Option "{0}" is invalid: \n'
                'For literal strings use "SECRET_NAME=VALUE" format'
                .format(literal),
                fg='red'),
            err=True)


def parse_secret_from_path(path):
    """Parse a file path into a secret dict.

    :param path: Path of the file containing secret
    :returns secret: Dictionary in the format suitable for sending
     via http request.
    """
    try:
        with open(os.path.expanduser(path), 'rb') as file:
            file_name = os.path.basename(path)
            secret = {
                file_name: {
                    'value': base64.b64encode(
                        file.read()).decode('utf-8'),
                    'type': 'file'
                }
            }
        return secret
    except FileNotFoundError as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        click.echo(
            click.style(
                'File {0} could not be uploaded: {0} does not exist.'
                .format(path),
                fg='red'),
            err=True)


def get_api_url():
    """Obtain REANA server API URL."""
    return os.environ.get('REANA_SERVER_URL')



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client debugging commands."""

import logging
import traceback

import click
from reana_client.cli.utils import check_connection
from reana_client.version import __version__


@click.group(help='Configuration commands')
def configuration_group():
    """Configuration commands."""
    pass


@configuration_group.command('ping')
@click.pass_context
@check_connection
def ping(ctx):  # noqa: D301
    """Check connection to REANA server.

    The `ping` command allows to test connection to REANA server.

    Examples: \n
    \t $ reana-client ping
    """
    try:
        from reana_client.api.client import ping as rs_ping
        from reana_client.utils import get_api_url
        logging.info('Connecting to {0}'.format(get_api_url()))
        response = rs_ping()
        click.echo(click.style('Connected to {0} - Server is running.'
                               .format(get_api_url()), fg='green'))
        logging.debug('Server response:\n{}'.format(response))

    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        error_msg = ('Could not connect to the selected REANA cluster'
                     'server at {0}.'.format(get_api_url()))
        click.echo(click.style(error_msg, fg='red'), err=True)


@configuration_group.command('version')
@click.pass_context
def version(ctx):  # noqa: D301
    """Show version.

    The `version` command shows REANA client version.

    Examples: \n
    \t $ reana-client version
    """
    click.echo(__version__)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA command line interface client."""
import logging
import os
import sys

import click
from urllib3 import disable_warnings

from reana_client.cli import workflow, files, ping, secrets

DEBUG_LOG_FORMAT = '[%(asctime)s] p%(process)s ' \
                   '{%(pathname)s:%(lineno)d} ' \
                   '%(levelname)s - %(message)s'

LOG_FORMAT = '[%(levelname)s] %(message)s'


class Config(object):
    """Configuration object to share across commands."""

    def __init__(self):
        """Initialize config variables."""
        self.reana_server_url = os.getenv('REANA_SERVER_URL', None)


class ReanaCLI(click.Group):
    """REANA command line interface."""

    cmd_groups = [ping.configuration_group,
                  workflow.workflow_management_group,
                  workflow.workflow_execution_group,
                  workflow.interactive_group,
                  files.files_group,
                  secrets.secrets_group]

    def __init__(self, name=None, commands=None, **attrs):
        """Initialize REANA client commands."""
        disable_warnings()
        click.Group.__init__(self, name, **attrs)
        for group in ReanaCLI.cmd_groups:
            for cmd in group.commands.items():
                self.add_command(cmd=cmd[1], name=cmd[0])

    def format_commands(self, ctx, formatter):
        """Overides default click cmd display."""
        if ReanaCLI.cmd_groups:
            max_cmd_length = \
                len(max([max(name) for name in self.list_commands(ctx)]))
            limit = formatter.width - 6 - max_cmd_length
            rows = []
            for group in ReanaCLI.cmd_groups:
                item = {'rows': []}
                item['group_help'] = group.get_short_help_str(limit)
                for command in sorted(group.commands.items()):
                    if command[1] is None:
                        continue
                    if command[1].hidden:
                        continue
                    command_help = command[1].get_short_help_str(limit)
                    item['rows'].append((command[0], command_help))
                rows.append(item)
            for item in rows:
                with formatter.section(item['group_help']):
                    formatter.write_dl(item['rows'])


@click.command(cls=ReanaCLI)
@click.option(
    '--loglevel',
    '-l',
    help='Sets log level',
    type=click.Choice(['DEBUG', 'INFO', 'WARNING']),
    default='WARNING')
@click.pass_context
@click.pass_obj
def cli(obj, ctx, loglevel):
    """REANA client for interacting with REANA server."""
    logging.basicConfig(
        format=DEBUG_LOG_FORMAT if loglevel == 'DEBUG' else LOG_FORMAT,
        stream=sys.stderr,
        level=loglevel)
    ctx.obj = obj or Config()
